# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("pdh.dll")
prototypes = \
    {
        #
        'PdhGetDllVersion': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="PDH_DLL_VERSION"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["lpdwVersion"]),
        #
        'PdhOpenQueryW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "dwUserData", "phQuery"]),
        #
        'PdhOpenQueryA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "dwUserData", "phQuery"]),
        #
        'PdhAddCounterW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hQuery", "szFullCounterPath", "dwUserData", "phCounter"]),
        #
        'PdhAddCounterA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hQuery", "szFullCounterPath", "dwUserData", "phCounter"]),
        #
        'PdhAddEnglishCounterW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hQuery", "szFullCounterPath", "dwUserData", "phCounter"]),
        #
        'PdhAddEnglishCounterA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hQuery", "szFullCounterPath", "dwUserData", "phCounter"]),
        #
        'PdhCollectQueryDataWithTime': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hQuery", "pllTimeStamp"]),
        #
        'PdhValidatePathExW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szFullPathBuffer"]),
        #
        'PdhValidatePathExA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szFullPathBuffer"]),
        #
        'PdhRemoveCounter': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter"]),
        #
        'PdhCollectQueryData': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hQuery"]),
        #
        'PdhCloseQuery': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hQuery"]),
        #
        'PdhGetFormattedCounterValue': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="PDH_FMT"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_FMT_COUNTERVALUE", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "dwFormat", "lpdwType", "pValue"]),
        #
        'PdhGetFormattedCounterArrayA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="PDH_FMT"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_FMT_COUNTERVALUE_ITEM_A", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "dwFormat", "lpdwBufferSize", "lpdwItemCount", "ItemBuffer"]),
        #
        'PdhGetFormattedCounterArrayW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="PDH_FMT"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_FMT_COUNTERVALUE_ITEM_W", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "dwFormat", "lpdwBufferSize", "lpdwItemCount", "ItemBuffer"]),
        #
        'PdhGetRawCounterValue': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_RAW_COUNTER", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "lpdwType", "pValue"]),
        #
        'PdhGetRawCounterArrayA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_RAW_COUNTER_ITEM_A", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "lpdwBufferSize", "lpdwItemCount", "ItemBuffer"]),
        #
        'PdhGetRawCounterArrayW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_RAW_COUNTER_ITEM_W", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "lpdwBufferSize", "lpdwItemCount", "ItemBuffer"]),
        #
        'PdhCalculateCounterFromRawValue': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="PDH_FMT"), SimTypePointer(SimTypeRef("PDH_RAW_COUNTER", SimStruct), offset=0), SimTypePointer(SimTypeRef("PDH_RAW_COUNTER", SimStruct), offset=0), SimTypePointer(SimTypeRef("PDH_FMT_COUNTERVALUE", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "dwFormat", "rawValue1", "rawValue2", "fmtValue"]),
        #
        'PdhComputeCounterStatistics': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="PDH_FMT"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("PDH_RAW_COUNTER", SimStruct), offset=0), SimTypePointer(SimTypeRef("PDH_STATISTICS", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "dwFormat", "dwFirstEntry", "dwNumEntries", "lpRawValueArray", "data"]),
        #
        'PdhGetCounterInfoW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_COUNTER_INFO_W", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "bRetrieveExplainText", "pdwBufferSize", "lpBuffer"]),
        #
        'PdhGetCounterInfoA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_COUNTER_INFO_A", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "bRetrieveExplainText", "pdwBufferSize", "lpBuffer"]),
        #
        'PdhSetCounterScaleFactor': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "lFactor"]),
        #
        'PdhConnectMachineW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szMachineName"]),
        #
        'PdhConnectMachineA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szMachineName"]),
        #
        'PdhEnumMachinesW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "mszMachineList", "pcchBufferSize"]),
        #
        'PdhEnumMachinesA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "mszMachineList", "pcchBufferSize"]),
        #
        'PdhEnumObjectsW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="PERF_DETAIL"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "szMachineName", "mszObjectList", "pcchBufferSize", "dwDetailLevel", "bRefresh"]),
        #
        'PdhEnumObjectsA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="PERF_DETAIL"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "szMachineName", "mszObjectList", "pcchBufferSize", "dwDetailLevel", "bRefresh"]),
        #
        'PdhEnumObjectItemsW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="PERF_DETAIL"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "szMachineName", "szObjectName", "mszCounterList", "pcchCounterListLength", "mszInstanceList", "pcchInstanceListLength", "dwDetailLevel", "dwFlags"]),
        #
        'PdhEnumObjectItemsA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="PERF_DETAIL"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "szMachineName", "szObjectName", "mszCounterList", "pcchCounterListLength", "mszInstanceList", "pcchInstanceListLength", "dwDetailLevel", "dwFlags"]),
        #
        'PdhMakeCounterPathW': SimTypeFunction([SimTypePointer(SimTypeRef("PDH_COUNTER_PATH_ELEMENTS_W", SimStruct), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="PDH_PATH_FLAGS")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pCounterPathElements", "szFullPathBuffer", "pcchBufferSize", "dwFlags"]),
        #
        'PdhMakeCounterPathA': SimTypeFunction([SimTypePointer(SimTypeRef("PDH_COUNTER_PATH_ELEMENTS_A", SimStruct), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="PDH_PATH_FLAGS")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pCounterPathElements", "szFullPathBuffer", "pcchBufferSize", "dwFlags"]),
        #
        'PdhParseCounterPathW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("PDH_COUNTER_PATH_ELEMENTS_W", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["szFullPathBuffer", "pCounterPathElements", "pdwBufferSize", "dwFlags"]),
        #
        'PdhParseCounterPathA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeRef("PDH_COUNTER_PATH_ELEMENTS_A", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["szFullPathBuffer", "pCounterPathElements", "pdwBufferSize", "dwFlags"]),
        #
        'PdhParseInstanceNameW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szInstanceString", "szInstanceName", "pcchInstanceNameLength", "szParentName", "pcchParentNameLength", "lpIndex"]),
        #
        'PdhParseInstanceNameA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szInstanceString", "szInstanceName", "pcchInstanceNameLength", "szParentName", "pcchParentNameLength", "lpIndex"]),
        #
        'PdhValidatePathW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szFullPathBuffer"]),
        #
        'PdhValidatePathA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szFullPathBuffer"]),
        #
        'PdhGetDefaultPerfObjectW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "szMachineName", "szDefaultObjectName", "pcchBufferSize"]),
        #
        'PdhGetDefaultPerfObjectA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "szMachineName", "szDefaultObjectName", "pcchBufferSize"]),
        #
        'PdhGetDefaultPerfCounterW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "szMachineName", "szObjectName", "szDefaultCounterName", "pcchBufferSize"]),
        #
        'PdhGetDefaultPerfCounterA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "szMachineName", "szObjectName", "szDefaultCounterName", "pcchBufferSize"]),
        #
        'PdhBrowseCountersW': SimTypeFunction([SimTypePointer(SimTypeRef("PDH_BROWSE_DLG_CONFIG_W", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pBrowseDlgData"]),
        #
        'PdhBrowseCountersA': SimTypeFunction([SimTypePointer(SimTypeRef("PDH_BROWSE_DLG_CONFIG_A", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pBrowseDlgData"]),
        #
        'PdhExpandCounterPathW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szWildCardPath", "mszExpandedPathList", "pcchPathListLength"]),
        #
        'PdhExpandCounterPathA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szWildCardPath", "mszExpandedPathList", "pcchPathListLength"]),
        #
        'PdhLookupPerfNameByIndexW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szMachineName", "dwNameIndex", "szNameBuffer", "pcchNameBufferSize"]),
        #
        'PdhLookupPerfNameByIndexA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szMachineName", "dwNameIndex", "szNameBuffer", "pcchNameBufferSize"]),
        #
        'PdhLookupPerfIndexByNameW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szMachineName", "szNameBuffer", "pdwIndex"]),
        #
        'PdhLookupPerfIndexByNameA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szMachineName", "szNameBuffer", "pdwIndex"]),
        #
        'PdhExpandWildCardPathA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "szWildCardPath", "mszExpandedPathList", "pcchPathListLength", "dwFlags"]),
        #
        'PdhExpandWildCardPathW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "szWildCardPath", "mszExpandedPathList", "pcchPathListLength", "dwFlags"]),
        #
        'PdhOpenLogW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="PDH_LOG"), SimTypePointer(SimTypeInt(signed=False, label="PDH_LOG_TYPE"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szLogFileName", "dwAccessFlags", "lpdwLogType", "hQuery", "dwMaxSize", "szUserCaption", "phLog"]),
        #
        'PdhOpenLogA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="PDH_LOG"), SimTypePointer(SimTypeInt(signed=False, label="PDH_LOG_TYPE"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szLogFileName", "dwAccessFlags", "lpdwLogType", "hQuery", "dwMaxSize", "szUserCaption", "phLog"]),
        #
        'PdhUpdateLogW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hLog", "szUserString"]),
        #
        'PdhUpdateLogA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hLog", "szUserString"]),
        #
        'PdhUpdateLogFileCatalog': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hLog"]),
        #
        'PdhGetLogFileSize': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hLog", "llSize"]),
        #
        'PdhCloseLog': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hLog", "dwFlags"]),
        #
        'PdhSelectDataSourceW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="PDH_SELECT_DATA_SOURCE_FLAGS"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hWndOwner", "dwFlags", "szDataSource", "pcchBufferLength"]),
        #
        'PdhSelectDataSourceA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="PDH_SELECT_DATA_SOURCE_FLAGS"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hWndOwner", "dwFlags", "szDataSource", "pcchBufferLength"]),
        #
        'PdhIsRealTimeQuery': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hQuery"]),
        #
        'PdhSetQueryTimeRange': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeRef("PDH_TIME_INFO", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hQuery", "pInfo"]),
        #
        'PdhGetDataSourceTimeRangeW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_TIME_INFO", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "pdwNumEntries", "pInfo", "pdwBufferSize"]),
        #
        'PdhGetDataSourceTimeRangeA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_TIME_INFO", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "pdwNumEntries", "pInfo", "pdwBufferSize"]),
        #
        'PdhCollectQueryDataEx': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hQuery", "dwIntervalTime", "hNewDataEvent"]),
        #
        'PdhFormatFromRawValue': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="PDH_FMT"), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0), SimTypePointer(SimTypeRef("PDH_RAW_COUNTER", SimStruct), offset=0), SimTypePointer(SimTypeRef("PDH_RAW_COUNTER", SimStruct), offset=0), SimTypePointer(SimTypeRef("PDH_FMT_COUNTERVALUE", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwCounterType", "dwFormat", "pTimeBase", "pRawValue1", "pRawValue2", "pFmtValue"]),
        #
        'PdhGetCounterTimeBase': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeLongLong(signed=True, label="Int64"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hCounter", "pTimeBase"]),
        #
        'PdhReadRawLogRecord': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeRef("FILETIME", SimStruct), SimTypePointer(SimTypeRef("PDH_RAW_LOG_RECORD", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hLog", "ftRecord", "pRawLogRecord", "pdwBufferLength"]),
        #
        'PdhSetDefaultRealTimeDataSource': SimTypeFunction([SimTypeInt(signed=False, label="REAL_TIME_DATA_SOURCE_ID_FLAGS")], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwDataSourceId"]),
        #
        'PdhBindInputDataSourceW': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["phDataSource", "LogFileNameList"]),
        #
        'PdhBindInputDataSourceA': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["phDataSource", "LogFileNameList"]),
        #
        'PdhOpenQueryH': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "dwUserData", "phQuery"]),
        #
        'PdhEnumMachinesHW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "mszMachineList", "pcchBufferSize"]),
        #
        'PdhEnumMachinesHA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "mszMachineList", "pcchBufferSize"]),
        #
        'PdhEnumObjectsHW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="PERF_DETAIL"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szMachineName", "mszObjectList", "pcchBufferSize", "dwDetailLevel", "bRefresh"]),
        #
        'PdhEnumObjectsHA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="PERF_DETAIL"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szMachineName", "mszObjectList", "pcchBufferSize", "dwDetailLevel", "bRefresh"]),
        #
        'PdhEnumObjectItemsHW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="PERF_DETAIL"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szMachineName", "szObjectName", "mszCounterList", "pcchCounterListLength", "mszInstanceList", "pcchInstanceListLength", "dwDetailLevel", "dwFlags"]),
        #
        'PdhEnumObjectItemsHA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="PERF_DETAIL"), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szMachineName", "szObjectName", "mszCounterList", "pcchCounterListLength", "mszInstanceList", "pcchInstanceListLength", "dwDetailLevel", "dwFlags"]),
        #
        'PdhExpandWildCardPathHW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szWildCardPath", "mszExpandedPathList", "pcchPathListLength", "dwFlags"]),
        #
        'PdhExpandWildCardPathHA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szWildCardPath", "mszExpandedPathList", "pcchPathListLength", "dwFlags"]),
        #
        'PdhGetDataSourceTimeRangeH': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeRef("PDH_TIME_INFO", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "pdwNumEntries", "pInfo", "pdwBufferSize"]),
        #
        'PdhGetDefaultPerfObjectHW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szMachineName", "szDefaultObjectName", "pcchBufferSize"]),
        #
        'PdhGetDefaultPerfObjectHA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szMachineName", "szDefaultObjectName", "pcchBufferSize"]),
        #
        'PdhGetDefaultPerfCounterHW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szMachineName", "szObjectName", "szDefaultCounterName", "pcchBufferSize"]),
        #
        'PdhGetDefaultPerfCounterHA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hDataSource", "szMachineName", "szObjectName", "szDefaultCounterName", "pcchBufferSize"]),
        #
        'PdhBrowseCountersHW': SimTypeFunction([SimTypePointer(SimTypeRef("PDH_BROWSE_DLG_CONFIG_HW", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pBrowseDlgData"]),
        #
        'PdhBrowseCountersHA': SimTypeFunction([SimTypePointer(SimTypeRef("PDH_BROWSE_DLG_CONFIG_HA", SimStruct), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pBrowseDlgData"]),
        #
        'PdhVerifySQLDBW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource"]),
        #
        'PdhVerifySQLDBA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource"]),
        #
        'PdhCreateSQLTablesW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource"]),
        #
        'PdhCreateSQLTablesA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource"]),
        #
        'PdhEnumLogSetNamesW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "mszDataSetNameList", "pcchBufferLength"]),
        #
        'PdhEnumLogSetNamesA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["szDataSource", "mszDataSetNameList", "pcchBufferLength"]),
        #
        'PdhGetLogSetGUID': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hLog", "pGuid", "pRunId"]),
        #
        'PdhSetLogSetRunID': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=False, label="UInt32"), arg_names=["hLog", "RunId"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations

import logging
from enum import Enum

from angr.engines.concrete import ConcreteEngine, HeavyConcreteState
from angr.errors import AngrError


log = logging.getLogger(name=__name__)


class EmulatorException(AngrError):
    """Base class for exceptions raised by the Emulator."""


class EngineException(EmulatorException):
    """Exception raised when the emulator encounters an unhandlable error in the engine."""


class StateDivergedException(EmulatorException):
    """Exception raised when an engine returns multiple successors."""


class EmulatorStopReason(Enum):
    """
    Enum representing the reason for stopping the emulator.
    """

    INSTRUCTION_LIMIT = "instruction_limit"
    BREAKPOINT = "breakpoint"
    NO_SUCCESSORS = "no_successors"
    MEMORY_ERROR = "memory_error"
    FAILURE = "failure"
    EXIT = "exit"


class Emulator:
    """
    Emulator is a utility that adapts an angr `ConcreteEngine` to a more
    user-friendly interface for concrete execution. It only supports concrete
    execution and requires a ConcreteEngine.

    Saftey: This class is not thread-safe. It should only be used in a
    single-threaded context. It can be safely shared between multiple threads,
    provided that only one thread is using it at a time.
    """

    _engine: ConcreteEngine
    _state: HeavyConcreteState

    def __init__(self, engine: ConcreteEngine, init_state: HeavyConcreteState):
        """
        :param engine: The `ConcreteEngine` to use for emulation.
        :param init_state: The initial state to use for emulation.
        """
        self._engine = engine
        self._state = init_state

    @property
    def state(self) -> HeavyConcreteState:
        """
        The current state of the emulator.
        """
        return self._state

    @property
    def breakpoints(self) -> set[int]:
        """
        The set of currently set breakpoints.
        """
        return self._engine.get_breakpoints()

    def add_breakpoint(self, addr: int) -> None:
        """
        Add a breakpoint at the given address.

        :param addr: The address to set the breakpoint at.
        """
        self._engine.add_breakpoint(addr)

    def remove_breakpoint(self, addr: int) -> None:
        """
        Remove a breakpoint at the given address, if present.

        :param addr: The address to remove the breakpoint from.
        """
        self._engine.remove_breakpoint(addr)

    def run(self, num_inst: int | None = None) -> EmulatorStopReason:
        """
        Execute the emulator.
        """
        completed_engine_execs = 0
        num_inst_executed: int = 0
        while self._state.history.jumpkind != "Ijk_Exit":
            # Check if there is a breakpoint at the current address
            if completed_engine_execs > 0 and self._state.addr in self._engine.get_breakpoints():
                return EmulatorStopReason.BREAKPOINT

            # Check if we've already executed the requested number of instructions
            if num_inst is not None and num_inst_executed >= num_inst:
                return EmulatorStopReason.INSTRUCTION_LIMIT

            # Calculate remaining instructions for this engine execution
            remaining_inst: int | None = None
            if num_inst is not None:
                remaining_inst = num_inst - num_inst_executed

            # Run the engine to get successors
            try:
                successors = self._engine.process(self._state, num_inst=remaining_inst)
            except EngineException as e:
                raise EngineException(f"Engine encountered an error: {e}") from e

            # Handle cases with an unexpected number of successors
            if len(successors.successors) == 0:
                return EmulatorStopReason.NO_SUCCESSORS
            if len(successors.successors) > 1:
                log.warning("Concrete engine returned multiple successors")

            # Set the state before raising further exceptions
            self._state = successors.successors[0]

            # Track the number of instructions executed using the state's history
            if self._state.history.recent_instruction_count > 0:
                num_inst_executed += self._state.history.recent_instruction_count

            if successors.successors[0].history.jumpkind == "Ijk_SigSEGV":
                return EmulatorStopReason.MEMORY_ERROR

            completed_engine_execs += 1

        return EmulatorStopReason.EXIT


__all__ = (
    "Emulator",
    "EmulatorException",
    "EmulatorStopReason",
    "EngineException",
    "StateDivergedException",
)



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("prntvpt.dll")
prototypes = \
    {
        #
        'PTQuerySchemaVersionSupport': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszPrinterName", "pMaxVersion"]),
        #
        'PTOpenProvider': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszPrinterName", "dwVersion", "phProvider"]),
        #
        'PTOpenProviderEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszPrinterName", "dwMaxVersion", "dwPrefVersion", "phProvider", "pUsedVersion"]),
        #
        'PTCloseProvider': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider"]),
        #
        'PTReleaseMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pBuffer"]),
        #
        'PTGetPrintCapabilities': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeBottom(label="IStream"), SimTypeBottom(label="IStream"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "pPrintTicket", "pCapabilities", "pbstrErrorMessage"]),
        #
        'PTGetPrintDeviceCapabilities': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeBottom(label="IStream"), SimTypeBottom(label="IStream"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "pPrintTicket", "pDeviceCapabilities", "pbstrErrorMessage"]),
        #
        'PTGetPrintDeviceResources': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IStream"), SimTypeBottom(label="IStream"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "pszLocaleName", "pPrintTicket", "pDeviceResources", "pbstrErrorMessage"]),
        #
        'PTMergeAndValidatePrintTicket': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeBottom(label="IStream"), SimTypeBottom(label="IStream"), SimTypeInt(signed=False, label="EPrintTicketScope"), SimTypeBottom(label="IStream"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "pBaseTicket", "pDeltaTicket", "scope", "pResultTicket", "pbstrErrorMessage"]),
        #
        'PTConvertPrintTicketToDevMode': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeBottom(label="IStream"), SimTypeInt(signed=False, label="EDefaultDevmodeType"), SimTypeInt(signed=False, label="EPrintTicketScope"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeRef("DEVMODEA", SimStruct), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "pPrintTicket", "baseDevmodeType", "scope", "pcbDevmode", "ppDevmode", "pbstrErrorMessage"]),
        #
        'PTConvertDevModeToPrintTicket': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeRef("DEVMODEA", SimStruct), offset=0), SimTypeInt(signed=False, label="EPrintTicketScope"), SimTypeBottom(label="IStream")], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "cbDevmode", "pDevmode", "scope", "pPrintTicket"]),
    }

lib.set_prototypes(prototypes)



from __future__ import annotations
from typing import TYPE_CHECKING, Any
import logging

import claripy

from angr.code_location import CodeLocation
from angr.project import Project
from angr.analyses.propagator.vex_vars import VEXReg
from .propagator_utils import PropagatorLoadCallback

if TYPE_CHECKING:
    from angr import SimState
    from angr.knowledge_plugins import Function


l = logging.getLogger(name=__name__)


class ConstantValueManager:
    """
    Manages the loading of registers who hold constant values.
    """

    __slots__ = (
        "func",
        "indirect_jump_addr",
        "kb",
        "mapping",
        "project",
    )

    def __init__(self, project: Project, kb, func: Function, ij_addr: int):
        self.project = project
        self.kb = kb
        self.func = func
        self.indirect_jump_addr = ij_addr

        self.mapping: dict[Any, dict[Any, claripy.ast.Base]] | None = None

    def reg_read_callback(self, state: SimState):
        if self.mapping is None:
            self._build_mapping()
            assert self.mapping is not None

        codeloc = CodeLocation(state.scratch.bbl_addr, state.scratch.stmt_idx, ins_addr=state.scratch.ins_addr)
        if codeloc in self.mapping:
            reg_read_offset = state.inspect.reg_read_offset
            if isinstance(reg_read_offset, claripy.ast.BV) and reg_read_offset.op == "BVV":
                reg_read_offset = reg_read_offset.args[0]
            variable = VEXReg(reg_read_offset, state.inspect.reg_read_length)
            if variable in self.mapping[codeloc]:
                v = self.mapping[codeloc][variable]
                if isinstance(v, int):
                    v = claripy.BVV(v, state.inspect.reg_read_length * state.arch.byte_width)
                state.inspect.reg_read_expr = v

    def _build_mapping(self):
        # constant propagation
        l.debug("JumpTable: Propagating for %r at %#x.", self.func, self.indirect_jump_addr)

        # determine blocks to run FCP on

        # - include at most three levels of superblock successors from the entrypoint
        self.mapping = {}
        startpoint = self.func.startpoint
        if startpoint is None:
            return

        blocks = set()
        succ_and_levels = [(startpoint, 0)]
        while succ_and_levels:
            new_succs = []
            for node, level in succ_and_levels:
                if node in blocks:
                    continue
                blocks.add(node)
                if node.addr == self.indirect_jump_addr:
                    # stop at the indirect jump block
                    continue
                for _, succ, data in self.func.graph.out_edges(node, data=True):
                    new_level = level if data.get("type") == "fake_return" else level + 1
                    if new_level <= 3:
                        new_succs.append((succ, new_level))
            succ_and_levels = new_succs

        # - include at most six levels of predecessors from the indirect jump block
        ij_block = self.func.get_node(self.indirect_jump_addr)
        preds = [ij_block]
        for _ in range(6):
            new_preds = []
            for node in preds:
                if node in blocks:
                    continue
                blocks.add(node)
                new_preds += list(self.func.graph.predecessors(node))
            preds = new_preds
            if not preds:
                break

        prop = self.project.analyses.FastConstantPropagation(
            self.func,
            blocks=blocks,
            vex_cross_insn_opt=True,
            load_callback=PropagatorLoadCallback(self.project).propagator_load_callback,
        )
        self.mapping = prop.replacements



from __future__ import annotations

import logging

import claripy

from angr.procedures.java import JavaSimProcedure
from angr.engines.soot.values.strref import SimSootValue_StringRef

l = logging.getLogger(name=__name__)


class GetSimpleName(JavaSimProcedure):
    __provides__ = (("java.lang.Class", "getSimpleName()"),)

    def run(self, this):  # pylint: disable=arguments-differ
        class_simple_name = this.type.split(".")[-1]
        return SimSootValue_StringRef.new_string(self.state, claripy.StringV(class_simple_name))



#!/usr/bin/env python3
# pylint: disable=missing-class-docstring,no-self-use,line-too-long,no-member
from __future__ import annotations

__package__ = __package__ or "tests"  # pylint:disable=redefined-builtin

import io
import os
import re
import sys
import unittest
from unittest import mock

import angr
from angr.__main__ import main
from angr.analyses.decompiler.utils import decompile_functions

from .common import bin_location


test_location = os.path.join(bin_location, "tests")


def run_cli(*args):
    with mock.patch("sys.argv", [sys.executable, *args]), mock.patch("sys.stdout", new=io.StringIO()) as fake_out:
        main()
        return fake_out.getvalue()


class TestCommandLineInterface(unittest.TestCase):
    def test_decompiling(self):
        bin_path = os.path.join(test_location, "x86_64", "decompiler", "sailr_motivating_example")
        f1 = "schedule_job"
        f2 = "main"

        # test a single function
        assert run_cli(bin_path, "decompile", "--functions", f1) == decompile_functions(bin_path, [f1]) + "\n"

        # test multiple functions
        assert run_cli(bin_path, "decompile", "--functions", f1, f2) == decompile_functions(bin_path, [f1, f2]) + "\n"

    def test_structuring(self):
        bin_path = os.path.join(test_location, "x86_64", "decompiler", "sailr_motivating_example")
        f1 = "schedule_job"

        dream_cli = run_cli(bin_path, "decompile", "--no-casts", "--functions", f1, "--structurer", "dream")
        sailr_cli = run_cli(bin_path, "decompile", "--functions", f1, "--structurer", "sailr")
        assert dream_cli.count("goto") == 0
        assert dream_cli != sailr_cli

    def test_base_addr_dec(self):
        bin_path = os.path.join(test_location, "x86_64", "decompiler", "sailr_motivating_example")
        f1 = "schedule_job"
        proj = angr.Project(bin_path, auto_load_libs=False)
        default_base_addr = proj.loader.main_object.min_addr
        f1_default_addr = proj.loader.find_symbol(f1).rebased_addr
        f1_offset = f1_default_addr - default_base_addr

        # function resolving is based on symbol
        sym_based_dec = run_cli(bin_path, "decompile", "--functions", f1, "--preset", "full")
        # function resolving is based on the address (with default angr loading)
        base_addr_dec = run_cli(bin_path, "decompile", "--functions", hex(f1_default_addr), "--preset", "full")
        # function resolving is based on the address (with base address specified)
        offset_dec = run_cli(
            bin_path, "--base-addr", "0x0", "decompile", "--functions", hex(f1_offset), "--preset", "full"
        )

        # since the externs can be unpredictable, we only check the function name down
        sym_based_dec = re.sub(r"extern .*;", "", sym_based_dec).lstrip("\n")
        base_addr_dec = re.sub(r"extern .*;", "", base_addr_dec).lstrip("\n")
        offset_dec = re.sub(r"extern .*;", "", offset_dec).lstrip("\n")

        # we must also normalize label names (since they are based on the address)
        sym_based_dec = re.sub(r"LABEL_[0-9a-fA-F]+", "LABEL_A", sym_based_dec)
        base_addr_dec = re.sub(r"LABEL_[0-9a-fA-F]+", "LABEL_A", base_addr_dec)
        offset_dec = re.sub(r"LABEL_[0-9a-fA-F]+", "LABEL_A", offset_dec)

        assert sym_based_dec == base_addr_dec == offset_dec

    def test_disassembly(self):
        bin_path = os.path.join(test_location, "x86_64", "fauxware")
        disasm = run_cli(bin_path, "disassemble")
        funcs = {
            "_init",
            "_start",
            "call_gmon_start",
            "__do_global_dtors_aux",
            "frame_dummy",
            "authenticate",
            "accepted",
            "rejected",
            "main",
            "__libc_csu_init",
            "__libc_csu_fini",
            "__do_global_ctors_aux",
            "_fini",
        }
        substrs = [f"{f}:" for f in funcs]
        substrs += "40071d  55              push    rbp"

        for s in substrs:
            assert s in disasm


if __name__ == "__main__":
    unittest.main()



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations

import archinfo

from angr.ailment import Block
from angr.ailment.statement import Statement, Call, Assignment
from angr.ailment.expression import Const, Register, VirtualVariable

from angr.analyses.decompiler.optimization_passes.optimization_pass import OptimizationPass, OptimizationPassStage
from angr.analyses.decompiler.optimization_passes import register_optimization_pass

WIN64_REG_ARGS = {
    archinfo.ArchAMD64().registers["rcx"][0],
    archinfo.ArchAMD64().registers["rdx"][0],
    archinfo.ArchAMD64().registers["r8"][0],
    archinfo.ArchAMD64().registers["r9"][0],
}


class StringObfType3Rewriter(OptimizationPass):
    """
    Type-3 optimization pass replaces deobfuscate_string calls with the deobfuscated strings, and then removes
    arguments on the stack.
    """

    ARCHES = ["X86", "AMD64"]
    PLATFORMS = ["windows"]
    STAGE = OptimizationPassStage.AFTER_MAKING_CALLSITES

    NAME = "Simplify Type 3 string deobfuscation calls"
    DESCRIPTION = "Simplify Type 3 string deobfuscation calls"
    stmt_classes = ()

    def __init__(self, func, **kwargs):
        super().__init__(func, **kwargs)

        self.analyze()

    def _check(self):
        if self.kb.obfuscations.type3_deobfuscated_strings:
            return True, None
        return False, None

    @staticmethod
    def is_call_or_call_assignment(stmt) -> bool:
        return isinstance(stmt, Call) or (isinstance(stmt, Assignment) and isinstance(stmt.src, Call))

    def _analyze(self, cache=None):

        # find all blocks with type-3 deobfuscation calls
        for block in list(self._graph):
            if not block.statements:
                continue
            last_stmt = block.statements[-1]
            if (
                self.is_call_or_call_assignment(last_stmt)
                and last_stmt.ins_addr in self.kb.obfuscations.type3_deobfuscated_strings
            ):
                new_block = self._process_block(
                    block, self.kb.obfuscations.type3_deobfuscated_strings[block.statements[-1].ins_addr]
                )
                if new_block is not None:
                    self._update_block(block, new_block)

    def _process_block(self, block: Block, deobf_content: bytes):
        # FIXME: This rewriter is very specific to the implementation of the deobfuscation scheme. we can make it more
        # generic when there are more cases available in the wild.

        # TODO: Support multiple blocks

        # replace the call
        old_stmt: Statement = block.statements[-1]
        str_id = self.kb.custom_strings.allocate(deobf_content)
        old_call: Call = old_stmt.src if isinstance(old_stmt, Assignment) else old_stmt
        new_call = Call(
            old_call.idx,
            "init_str",
            args=[
                old_call.args[0],
                Const(None, None, str_id, self.project.arch.bits, custom_string=True),
                Const(None, None, len(deobf_content), self.project.arch.bits),
            ],
            ret_expr=old_call.ret_expr,
            bits=old_call.bits,
            **old_call.tags,
        )
        if isinstance(old_stmt, Assignment):
            new_stmt = Assignment(old_stmt.idx, old_stmt.dst, new_call, **old_stmt.tags)
        else:
            new_stmt = new_call

        statements = [*block.statements[:-1], new_stmt]

        # remove N-2 continuous stack assignment
        if len(deobf_content) > 2:
            stack_offset_to_stmtid: dict[int, int] = {}
            for idx, stmt in enumerate(statements):
                if (
                    isinstance(stmt, Assignment)
                    and isinstance(stmt.dst, VirtualVariable)
                    and stmt.dst.was_stack
                    and isinstance(stmt.dst.stack_offset, int)
                    and isinstance(stmt.src, Const)
                    and stmt.src.value <= 0xFF
                ):
                    stack_offset_to_stmtid[stmt.dst.stack_offset] = idx
            sorted_offsets = sorted(stack_offset_to_stmtid)
            if sorted_offsets:
                spacing = 8  # FIXME: Make it adjustable
                distance = min(len(deobf_content) - 2, len(sorted_offsets) - 1)
                for start_idx in range(len(sorted_offsets) - distance):
                    if sorted_offsets[start_idx] + spacing * distance == sorted_offsets[start_idx + distance]:
                        # found them
                        # remove these statements
                        for i in range(start_idx, start_idx + distance + 1):
                            statements[stack_offset_to_stmtid[sorted_offsets[i]]] = None
                        break
                statements = [stmt for stmt in statements if stmt is not None]

        # remove writes to rdx, rcx, r8, and r9
        if self.project.arch.name == "AMD64":
            statements = [stmt for stmt in statements if not self._stmt_sets_win64_reg_arg(stmt)]

        # return the new block
        return block.copy(statements=statements)

    @staticmethod
    def _stmt_sets_win64_reg_arg(stmt) -> bool:
        return isinstance(stmt, Assignment) and isinstance(stmt.dst, Register) and stmt.dst.reg_offset in WIN64_REG_ARGS


register_optimization_pass(StringObfType3Rewriter, presets=["fast", "full"])



from __future__ import annotations
import angr


class Caller(angr.SimProcedure):
    """
    Caller stub. Creates a Ijk_Call exit to the specified function
    """

    def run(self, target_addr=None, target_cc=None):
        self.call(target_addr, [], "after_call", cc=target_cc, prototype="void x()")

    def after_call(self, target_addr=None, target_cc=None):
        pass



from __future__ import annotations
from math import gcd

from angr.ailment.expression import BinaryOp, UnaryOp, Const, Convert, StackBaseOffset

from angr.utils.bits import sign_extend
from .base import PeepholeOptimizationExprBase


class EagerEvaluation(PeepholeOptimizationExprBase):
    """
    Eagerly evaluates certain types of expressions.
    """

    __slots__ = ()

    NAME = "Eager expression evaluation"
    expr_classes = (BinaryOp, UnaryOp, Convert)

    def optimize(self, expr, **kwargs):  # type:ignore
        if isinstance(expr, BinaryOp):
            return self._optimize_binaryop(expr)
        if isinstance(expr, Convert):
            return self._optimize_convert(expr)
        if isinstance(expr, UnaryOp):
            return self._optimize_unaryop(expr)
        return None

    @staticmethod
    def _optimize_binaryop(expr: BinaryOp):
        if expr.op == "Add":
            if (
                isinstance(expr.operands[0], Const)
                and isinstance(expr.operands[0].value, int)
                and isinstance(expr.operands[1], Const)
                and isinstance(expr.operands[1].value, int)
            ):
                mask = (1 << expr.bits) - 1
                return Const(
                    expr.idx, None, (expr.operands[0].value + expr.operands[1].value) & mask, expr.bits, **expr.tags
                )
            if isinstance(expr.operands[1], Const) and expr.operands[1].value == 0:
                return expr.operands[0]
            if (
                isinstance(expr.operands[1], Const)
                and isinstance(expr.operands[0], BinaryOp)
                and isinstance(expr.operands[0].operands[1], Const)
            ):
                left = expr.operands[0]
                inner_expr, const_0 = left.operands
                const_1 = expr.operands[1]
                if left.op == "Add":
                    return BinaryOp(
                        left.idx,
                        "Add",
                        [inner_expr, Const(const_0.idx, None, const_0.value + const_1.value, const_0.bits)],
                        expr.signed,
                        **expr.tags,
                    )
                if left.op == "Sub":
                    return BinaryOp(
                        left.idx,
                        "Add",
                        [inner_expr, Const(const_0.idx, None, const_1.value - const_0.value, const_0.bits)],
                        expr.signed,
                        **expr.tags,
                    )
            op0, op1 = expr.operands
            if op0.likes(op1):
                # x + x => 2 * x
                count = Const(expr.idx, None, 2, op0.bits, **expr.tags)
                return BinaryOp(expr.idx, "Mul", [op0, count], expr.signed, **expr.tags)

            op0_is_mulconst = (
                isinstance(op0, BinaryOp)
                and op0.op == "Mul"
                and (isinstance(op0.operands[0], Const) or isinstance(op0.operands[1], Const))
            )
            op1_is_mulconst = (
                isinstance(op1, BinaryOp)
                and op1.op == "Mul"
                and (isinstance(op1.operands[0], Const) or isinstance(op1.operands[1], Const))
            )
            const0, x0 = None, None
            const1, x1 = None, None
            if op0_is_mulconst:
                if isinstance(op0.operands[0], Const):
                    const0, x0 = op0.operands
                elif isinstance(op0.operands[1], Const):
                    x0, const0 = op0.operands
            if op1_is_mulconst:
                if isinstance(op1.operands[0], Const):
                    const1, x1 = op1.operands
                elif isinstance(op1.operands[1], Const):
                    x1, const1 = op1.operands

            if op0_is_mulconst ^ op1_is_mulconst:
                if x0 is not None and const0 is not None and x0.likes(op1):
                    # x * A + x => (A + 1) * x
                    new_const = Const(const0.idx, None, const0.value + 1, const0.bits, **const0.tags)
                    return BinaryOp(expr.idx, "Mul", [x0, new_const], expr.signed, **expr.tags)
                if x1 is not None and const1 is not None and x1.likes(op0):
                    # x + x * A => (A + 1) * x
                    new_const = Const(const1.idx, None, const1.value + 1, const1.bits, **const1.tags)
                    return BinaryOp(expr.idx, "Mul", [x1, new_const], expr.signed, **expr.tags)
            elif op0_is_mulconst and op1_is_mulconst:
                assert x0 is not None and x1 is not None and const0 is not None and const1 is not None
                if x0.likes(x1):
                    # x * A + x * B => (A + B) * x
                    new_const = Const(const0.idx, None, const0.value + const1.value, const0.bits, **const0.tags)
                    return BinaryOp(expr.idx, "Mul", [x0, new_const], expr.signed, **expr.tags)

        elif expr.op == "Sub":
            if (
                isinstance(expr.operands[0], Const)
                and isinstance(expr.operands[0].value, int)
                and isinstance(expr.operands[1], Const)
                and isinstance(expr.operands[1].value, int)
            ):
                mask = (1 << expr.bits) - 1
                return Const(
                    expr.idx, None, (expr.operands[0].value - expr.operands[1].value) & mask, expr.bits, **expr.tags
                )
            if isinstance(expr.operands[1], Const) and expr.operands[1].is_int and expr.operands[1].sign_bit == 1:
                # x - (-A)  ==>  x + A
                assert isinstance(expr.operands[1].value, int)
                mask = (1 << expr.operands[1].bits) - 1
                complement = Const(
                    expr.operands[1].idx,
                    expr.operands[1].variable,
                    ((~expr.operands[1].value) + 1) & mask,
                    expr.operands[1].bits,
                    **expr.operands[1].tags,
                )
                return BinaryOp(
                    expr.idx,
                    "Add",
                    [expr.operands[0], complement],
                    expr.signed,
                    variable=expr.variable,
                    variable_offset=expr.variable_offset,
                    bits=expr.bits,
                    **expr.tags,
                )

            if (
                isinstance(expr.operands[1], Const)
                and isinstance(expr.operands[0], BinaryOp)
                and isinstance(expr.operands[0].operands[1], Const)
            ):
                left = expr.operands[0]
                inner_expr, const_0 = left.operands
                const_1 = expr.operands[1]
                if left.op == "Add":
                    return BinaryOp(
                        left.idx,
                        "Sub",
                        [inner_expr, Const(const_0.idx, None, const_1.value - const_0.value, const_0.bits)],
                        expr.signed,
                        **expr.tags,
                    )
                if left.op == "Sub":
                    return BinaryOp(
                        left.idx,
                        "Sub",
                        [inner_expr, Const(const_0.idx, None, const_0.value + const_1.value, const_0.bits)],
                        expr.signed,
                        **expr.tags,
                    )
            if isinstance(expr.operands[0], Const) and expr.operands[0].value == 0:
                return UnaryOp(expr.idx, "Neg", expr.operands[1], **expr.tags)

            r = EagerEvaluation._combine_like_terms(expr)
            if r is not None:
                return r

            if isinstance(expr.operands[0], StackBaseOffset) and isinstance(expr.operands[1], StackBaseOffset):
                assert isinstance(expr.operands[0].offset, int) and isinstance(expr.operands[1].offset, int)
                return Const(expr.idx, None, expr.operands[0].offset - expr.operands[1].offset, expr.bits, **expr.tags)

        elif expr.op == "And":
            op0, op1 = expr.operands
            if (
                isinstance(op0, Const)
                and isinstance(op0.value, int)
                and isinstance(op1, Const)
                and isinstance(op1.value, int)
            ):
                return Const(expr.idx, None, (op0.value & op1.value), expr.bits, **expr.tags)
            if isinstance(op1, Const) and op1.value == 0:
                return Const(expr.idx, None, 0, expr.bits, **expr.tags)

        elif expr.op == "Mul":
            if isinstance(expr.operands[1], Const) and expr.operands[1].value == 1:
                # x * 1 => x
                return expr.operands[0]
            if (
                isinstance(expr.operands[0], Const)
                and expr.operands[0].is_int
                and isinstance(expr.operands[1], Const)
                and expr.operands[1].is_int
            ):
                assert isinstance(expr.operands[0].value, int) and isinstance(expr.operands[1].value, int)
                # constant multiplication
                mask = (1 << expr.bits) - 1
                return Const(
                    expr.idx, None, (expr.operands[0].value * expr.operands[1].value) & mask, expr.bits, **expr.tags
                )
            if {type(expr.operands[0]), type(expr.operands[1])} == {BinaryOp, Const}:
                op0, op1 = expr.operands
                const_, x0 = (op0, op1) if isinstance(op0, Const) else (op1, op0)
                if x0.op == "Mul" and (isinstance(x0.operands[0], Const) or isinstance(x0.operands[1], Const)):
                    # (A * x) * C => (A * C) * x
                    if isinstance(x0.operands[0], Const):
                        const_x0, x = x0.operands[0], x0.operands[1]
                    else:
                        const_x0, x = x0.operands[1], x0.operands[0]
                    new_const = Const(const_.idx, None, const_.value * const_x0.value, const_.bits, **const_x0.tags)
                    return BinaryOp(expr.idx, "Mul", [x, new_const], expr.signed, bits=expr.bits, **expr.tags)

        elif (
            expr.op == "Div"
            and isinstance(expr.operands[1], Const)
            and isinstance(expr.operands[0], BinaryOp)
            and expr.operands[0].op == "Mul"
            and isinstance(expr.operands[0].operands[1], Const)
        ):
            expr0, const_0 = expr.operands
            const_1 = expr0.operands[1]
            if const_0.value != 0 and const_1.value != 0:
                gcd_ = gcd(const_0.value, const_1.value)
                if gcd_ != 1:
                    new_const_1 = Const(
                        const_1.idx, const_1.variable, const_1.value // gcd_, const_1.bits, **const_1.tags
                    )
                    new_const_0 = Const(
                        const_0.idx, const_0.variable, const_0.value // gcd_, const_0.bits, **const_0.tags
                    )
                    mul = BinaryOp(
                        expr0.idx,
                        "Mul",
                        (expr0.operands[0], new_const_1),
                        expr0.signed,
                        variable=expr0.variable,
                        variable_offset=expr0.variable_offset,
                        bits=expr0.bits,
                        **expr0.tags,
                    )
                    return BinaryOp(expr.idx, "Div", (mul, new_const_0), expr.signed, bits=expr.bits, **expr.tags)

        elif expr.op == "Mod":
            op0, op1 = expr.operands
            if (
                isinstance(op0, Const)
                and isinstance(op0.value, int)
                and isinstance(op1, Const)
                and isinstance(op1.value, int)
                and op1.value != 0
            ):
                return Const(expr.idx, None, op0.value % op1.value, expr.bits, **expr.tags)

        elif expr.op in {"Shr", "Sar"} and isinstance(expr.operands[1], Const):
            expr0, expr1 = expr.operands
            if isinstance(expr0, BinaryOp) and expr0.op == "Shr" and isinstance(expr0.operands[1], Const):
                # (a >> M) >> N  ==>  a >> (M + N)
                const_a = expr0.operands[1]
                const_b = expr1
                const = Const(const_b.idx, None, const_a.value + const_b.value, const_b.bits, **const_b.tags)
                return BinaryOp(expr.idx, expr.op, (expr0.operands[0], const), False, bits=expr.bits, **expr.tags)

            if isinstance(expr0, BinaryOp) and expr0.op == "Div" and isinstance(expr0.operands[1], Const):
                # (a / M0) >> M1  ==>  a / (M0 * 2 ** M1)
                const_m0 = expr0.operands[1]
                const_m1 = expr1
                const = Const(const_m0.idx, None, const_m0.value * 2**const_m1.value, const_m0.bits, **const_m0.tags)
                return BinaryOp(expr.idx, "Div", (expr0.operands[0], const), expr.signed, bits=expr.bits, **expr.tags)

            if isinstance(expr0, Const):
                const_a = expr0.value
                mask = (2**expr0.bits) - 1
                return Const(expr0.idx, None, (const_a >> expr1.value) & mask, expr0.bits, **expr0.tags)

            if expr.op == "Shr" and expr.operands[0].bits <= expr.operands[1].value:
                return Const(expr.idx, None, 0, expr.operands[0].bits, **expr.tags)

        elif expr.op == "Shl" and isinstance(expr.operands[1], Const):
            expr0, expr1 = expr.operands
            if isinstance(expr0, Const):
                const_a = expr0.value
                mask = (2**expr0.bits) - 1
                return Const(expr0.idx, None, (const_a << expr1.value) & mask, expr0.bits, **expr0.tags)

        elif expr.op == "Or":
            op0, op1 = expr.operands
            if (
                isinstance(op0, Const)
                and isinstance(op0.value, int)
                and isinstance(op1, Const)
                and isinstance(op1.value, int)
            ):
                return Const(expr.idx, None, expr.operands[0].value | expr.operands[1].value, expr.bits, **expr.tags)
            if isinstance(expr.operands[0], Const) and expr.operands[0].value == 0:
                return expr.operands[1]
            if isinstance(expr.operands[1], Const) and expr.operands[1].value == 0:
                return expr.operands[0]
            if isinstance(expr.operands[0], Const) and expr.operands[0].value == (1 << expr.operands[0].bits) - 1:
                return expr.operands[0]
            if isinstance(expr.operands[1], Const) and expr.operands[1].value == (1 << expr.operands[1].bits) - 1:
                return expr.operands[1]
            if expr.operands[0].likes(expr.operands[1]):
                return expr.operands[0]

        elif expr.op == "Xor":
            op0, op1 = expr.operands
            if (
                isinstance(op0, Const)
                and isinstance(op0.value, int)
                and isinstance(op1, Const)
                and isinstance(op1.value, int)
            ):
                return Const(expr.idx, None, expr.operands[0].value ^ expr.operands[1].value, expr.bits, **expr.tags)

        elif expr.op in {"CmpEQ", "CmpLE", "CmpGE"}:
            if expr.operands[0].likes(expr.operands[1]):
                # x == x => 1
                return Const(expr.idx, None, 1, 1, **expr.tags)
            if isinstance(expr.operands[0], Const) and isinstance(expr.operands[1], Const):
                if expr.op == "CmpEQ":
                    return Const(
                        expr.idx, None, 1 if expr.operands[0].value == expr.operands[1].value else 0, 1, **expr.tags
                    )
                if expr.op == "CmpLE":
                    return Const(
                        expr.idx, None, 1 if expr.operands[0].value <= expr.operands[1].value else 0, 1, **expr.tags
                    )
                if expr.op == "CmpGE":
                    return Const(
                        expr.idx, None, 1 if expr.operands[0].value >= expr.operands[1].value else 0, 1, **expr.tags
                    )

        elif expr.op in {"CmpNE", "CmpLT", "CmpGT"}:
            if expr.operands[0].likes(expr.operands[1]):
                # x != x => 0
                return Const(expr.idx, None, 0, 1, **expr.tags)
            if isinstance(expr.operands[0], Const) and isinstance(expr.operands[1], Const):
                if expr.op == "CmpNE":
                    return Const(
                        expr.idx, None, 1 if expr.operands[0].value != expr.operands[1].value else 0, 1, **expr.tags
                    )
                if expr.op == "CmpLT":
                    return Const(
                        expr.idx, None, 1 if expr.operands[0].value < expr.operands[1].value else 0, 1, **expr.tags
                    )
                if expr.op == "CmpGT":
                    return Const(
                        expr.idx, None, 1 if expr.operands[0].value > expr.operands[1].value else 0, 1, **expr.tags
                    )

        return None

    @staticmethod
    def _combine_like_terms(expr: BinaryOp) -> BinaryOp | None:
        """
        Combine like terms for binary operations.
        """

        op = expr.op
        assert op in {"Add", "Sub"}

        expr0, expr1 = expr.operands

        conv = None
        if isinstance(expr0, Convert) and expr0.from_bits < expr0.to_bits:
            conv = expr0.from_bits, expr0.to_bits, expr0.is_signed
            expr0 = expr0.operand

        if isinstance(expr0, BinaryOp) and expr0.op == "Mul" and isinstance(expr0.operands[1], Const):
            n = expr0.operands[0]

            if isinstance(n, Convert) and n.from_bits > n.to_bits:
                if conv is not None and (n.to_bits, n.from_bits, n.is_signed) != conv:
                    return None
                n = n.operand

            if n.likes(expr1):
                # (n * C) - n  ==>  (C - 1) * n
                coeff_0 = expr0.operands[1]
                coeff = Const(coeff_0.idx, None, coeff_0.value - 1, expr.bits, **coeff_0.tags)
                return BinaryOp(
                    expr.idx, "Mul", [n, coeff], expr.signed, variable=expr.variable, bits=expr.bits, **expr.tags
                )
            if isinstance(expr1, BinaryOp) and expr1.op == "Mul" and isinstance(expr.operands[1].operands[1], Const):
                n1 = expr.operands[1].operands[0]
                if n.likes(n1):
                    # (n * C) - (n1 * C1)  ==>  n * (C - C1)
                    coeff_0 = expr0.operands[1]
                    coeff_1 = expr1.operands[1]
                    coeff = Const(coeff_0.idx, None, coeff_0.value - coeff_1.value, expr.bits, **coeff_0.tags)
                    return BinaryOp(
                        expr.idx,
                        "Mul",
                        [n, coeff],
                        expr.signed,
                        variable=expr.variable,
                        bits=expr.bits,
                        **expr.tags,
                    )
        return None

    @staticmethod
    def _optimize_unaryop(expr: UnaryOp):
        if expr.op == "Neg" and isinstance(expr.operand, Const) and isinstance(expr.operand.value, int):
            const_a = expr.operand.value
            mask = (2**expr.bits) - 1
            return Const(expr.idx, None, (~const_a) & mask, expr.bits, **expr.tags)

        return None

    @staticmethod
    def _optimize_convert(expr: Convert):
        if (
            isinstance(expr.operand, Const)
            and expr.operand.is_int
            and expr.from_type == Convert.TYPE_INT
            and expr.to_type == Convert.TYPE_INT
            and expr.from_bits > expr.to_bits
        ):
            assert isinstance(expr.operand.value, int)
            # truncation
            mask = (1 << expr.to_bits) - 1
            v = expr.operand.value & mask
            return Const(expr.idx, expr.operand.variable, v, expr.to_bits, **expr.operand.tags)
        if (
            isinstance(expr.operand, Const)
            and expr.operand.is_int
            and expr.from_type == Convert.TYPE_INT
            and expr.to_type == Convert.TYPE_INT
            and expr.from_bits <= expr.to_bits
        ):
            assert isinstance(expr.operand.value, int)
            if expr.is_signed is False:
                # unsigned extension
                return Const(expr.idx, expr.operand.variable, expr.operand.value, expr.to_bits, **expr.operand.tags)
            # signed extension
            v = sign_extend(expr.operand.value, expr.to_bits)
            return Const(expr.idx, expr.operand.variable, v, expr.to_bits, **expr.operand.tags)
        return None



from __future__ import annotations
from .flirt_module import FlirtModule


class FlirtNode:
    """
    Describes a tree node in the FLIRT signature tree.
    """

    __slots__ = ("children", "length", "modules", "pattern")

    def __init__(self, children: list[FlirtNode], modules: list[FlirtModule], length: int, pattern: list[int]):
        self.children = children
        self.modules = modules
        self.length = length
        self.pattern = pattern

    @property
    def leaf(self) -> bool:
        return not self.children

    def __repr__(self) -> str:
        return f"<FlirtNode length={self.length} leaf={self.leaf}>"



# pylint:disable=line-too-long
from __future__ import annotations
import logging
from collections import OrderedDict

from angr.sim_type import SimTypeFunction, SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat, SimTypePointer, SimTypeChar, SimStruct, SimTypeArray, SimTypeBottom, SimUnion, SimTypeBool, SimTypeRef
from angr.calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from angr.procedures import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.type_collection_names = ["win32"]
lib.set_default_cc("X86", SimCCStdcall)
lib.set_default_cc("AMD64", SimCCMicrosoftAMD64)
lib.set_library_names("srpapi.dll")
prototypes = \
    {
        #
        'SrpCreateThreadNetworkContext': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeRef("HTHREAD_NETWORK_CONTEXT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["enterpriseId", "threadNetworkContext"]),
        #
        'SrpCloseThreadNetworkContext': SimTypeFunction([SimTypePointer(SimTypeRef("HTHREAD_NETWORK_CONTEXT", SimStruct), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["threadNetworkContext"]),
        #
        'SrpSetTokenEnterpriseId': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["tokenHandle", "enterpriseId"]),
        #
        'SrpGetEnterpriseIds': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["tokenHandle", "numberOfBytes", "enterpriseIds", "enterpriseIdCount"]),
        #
        'SrpEnablePermissiveModeFileEncryption': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["enterpriseId"]),
        #
        'SrpDisablePermissiveModeFileEncryption': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'SrpGetEnterprisePolicy': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="ENTERPRISE_DATA_POLICIES"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["tokenHandle", "policyFlags"]),
        #
        'SrpIsTokenService': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["TokenHandle", "IsTokenService"]),
        #
        'SrpDoesPolicyAllowAppExecution': SimTypeFunction([SimTypePointer(SimTypeRef("PACKAGE_ID", SimStruct), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["packageId", "isAllowed"]),
        #
        'SrpHostingInitialize': SimTypeFunction([SimTypeInt(signed=False, label="SRPHOSTING_VERSION"), SimTypeInt(signed=False, label="SRPHOSTING_TYPE"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["Version", "Type", "pvData", "cbData"]),
        #
        'SrpHostingTerminate': SimTypeFunction([SimTypeInt(signed=False, label="SRPHOSTING_TYPE")], SimTypeBottom(label="Void"), arg_names=["Type"]),
    }

lib.set_prototypes(prototypes)
